//
//  PQStuckPointEditerController.swift
//  PQSpeed
//
//  Created by ak on 2021/4/26.
//  Copyright © 2021 BytesFlow. All rights reserved.
//  功能：卡点音乐编辑界面

// 创建不同玩法的类型
enum createStickersModel {
    case createStickersModelPoint // 卡点
    case createStickersModelSpeed // 快慢速
    case createStickersModelOnlyMusic //仅配乐
}

import Foundation
import ObjectMapper
import RealmSwift
import UIKit

class PQStuckPointEditerController: PQBaseViewController {
    // 是否导出视频成功
    var isExportVideosSuccess: Bool = false
    // 是否请求卡点数据成功
    var isStuckPointDataSuccess: Bool = false
    // 是否同步音乐成功
    var isSynchroMusicInfoSuccess: Bool = false
    /// 当前所有的filter
    var filters: Array = Array<ImageProcessingOperation>.init()
    // 选中所有素材的的总时长 其中图片时长为1.5
    var selectedTotalDuration: Float64 = 0
    // 选择的总数
    var selectedDataCount: Int = 0
    // 选择的图片总数
    var selectedImageDataCount: Int = 0
    // 选中的素材数据
    var selectedPhotoData: [PQEditVisionTrackMaterialsModel]?
    // 选中的音乐数据
    var stuckPointMusicData: PQVoiceModel?
    // 保存所有段的所有贴纸，音频信息,用于播放器的渲染使用
    var projectModel: PQEditProjectModel = PQEditProjectModel()
    // 从草稿箱进入的项目数据
    var draftProjectModel: PQEditProjectModel?
    var mStickers: [PQEditVisionTrackMaterialsModel]?
    // 播放器的开始和结束时间,1，刚进界面使用推荐的开始结束时间，2，用户修改起结点时修改
    var playeTimeRange: CMTimeRange = CMTimeRange()

    // add by ak 是否是再创作模式
    var isReCreate: Bool = false
    // 最后一个选择的模式 BTN 用于还原选中状态
    var lastEditModelBtn: UIButton?
    
    // add by ak  最大、最小速度 有固定值和自定义，当快慢速下两个值都有效，当跳跃卡点只有maxSpeed有效
    var maxSpeed: Float = 0.0
    var minSpeed: Float = 0.0
    
    //快慢速最后一次选择的速度位置
    var lastSpeedSelectIndex:Int = 0
    //跳跃卡点最后一次选择的速度位置
    var lastJumpSpeedSelectIndex:Int = 0
    
    //当前选择的玩法模式
    var currentCreateStickersModel:createStickersModel = .createStickersModelSpeed
    
    //最终使用的卡点数据
    var finallyStuckPoints: Array = Array<Float>.init()
    //最终使用的音频时长
    var finallyUserAudioTime:Float = 0.0
    // 下一步
    lazy var nextBtn: UIButton = {
        let nextBtn = UIButton(type: .custom)
        nextBtn.frame = CGRect(x: cScreenWidth - 16 - cDefaultMargin * 6, y: cDevice_iPhoneStatusBarHei + (cDevice_iPhoneNavBarHei - cDefaultMargin * 3) / 2, width: cDefaultMargin * 6, height: cDefaultMargin * 3)
        nextBtn.setTitle("合成", for: .normal)
        nextBtn.titleLabel?.font = UIFont.systemFont(ofSize: 15, weight: .medium)
        nextBtn.addTarget(self, action: #selector(nextBtnClick(sender:)), for: .touchUpInside)
        nextBtn.backgroundColor = UIColor.hexColor(hexadecimal: PQBFConfig.shared.styleColor.rawValue)
        nextBtn.setTitleColor(UIColor.hexColor(hexadecimal: "#FFFFFF"), for: .normal)
        nextBtn.uxy_acceptEventInterval = 0.5
        nextBtn.addCorner(corner: 3)
        return nextBtn
    }()

    // 播放器显示 view
    lazy var playerView: PQGPUImagePlayerView = {
        let playerHeight = cScreenWidth
        let playerView = PQGPUImagePlayerView(frame: CGRect(x: 0, y: navHeadImageView?.frame.maxY ?? 0, width: playerHeight, height: playerHeight))
        playerView.backgroundColor = PQBFConfig.shared.styleBackGroundColor
        playerView.isShowLine = false
        playerView.showGaussianBlur = true
        playerView.renderViewOnClickHandle = {[weak self] in
            self?.musicEditBGView.pausePlayer()
        }
        playerView.playerEmptyView.isHidden = true
        return playerView
    }()

    /// 节奏选择视图
    lazy var sustomSwitchView: PQCustomSwitchView = {
        let sustomSwitchView = PQCustomSwitchView(frame: CGRect(x: 16, y: 0, width: 180, height: 30), titles: ["快节奏", "适中", "慢节奏"], defaultIndex: stuckPointMusicData?.speed ?? 2)
        sustomSwitchView.switchChangeHandle = { [weak self] sender in
            // 改变速率
            self?.stuckPointMusicData?.speed = sender.tag
            self?.projectModel.sData?.getBGMSession()?.sectionTimeline?.audioTrack?.audioTrackMaterials.first?.bgmInfo?.rhythmMusicSpeed = sender.tag
            // 播放前先暂停
//            self?.playerView.stop()
            // 开始播放
            self?.settingPlayerView()
            // 点击上报：选择节奏
            PQEventTrackViewModel.baseReportUpload(businessType: .bt_buttonClick, objectType: .ot_click_selectRhythm, pageSource: .sp_stuck_previewSyncedUp, extParams: ["rhythmNumber": sender.tag], remindmsg: "点击上报：选择节奏")
        }
        return sustomSwitchView
    }()

    /// 裁剪视图
    lazy var stuckPointCuttingView: PQStuckPointCuttingView = {
         
        let stuckPointCuttingView = PQStuckPointCuttingView(frame: CGRect(x: 0, y: optionlineView.frame.minY - 85 - 28, width: view.frame.width, height: 80), duration: CGFloat(Float(stuckPointMusicData?.duration ?? "0") ?? 0), startTime: CGFloat(stuckPointMusicData?.startTime ?? 0), endTime: CGFloat(stuckPointMusicData?.endTime ?? 0))
        /// 裁剪进度回调
        stuckPointCuttingView.videoRangeDidChanged = { [weak self] startTime, endTime in
            BFLog(message: "裁剪返回--startTime = \(startTime),endTime = \(endTime)")
        }
        /// 播放进度回调
        stuckPointCuttingView.videoProgressDidChanged = { [weak self] progress in
            BFLog(message: "进度更新返回--progress = \(progress) \(String(describing: self?.playerView.mPlayeTimeRange))")
        }
        /// 拖缀结束的回调 type - 1-拖动左边裁剪结束 2--拖动右边裁剪结束 3-进度条拖动结束 4-滑动结束
        stuckPointCuttingView.videoDidEndDragging = { [weak self] type, startTime, endTime, progress in
            BFLog(message: "拖拽结束返回--type = \(type),startTime = \(startTime),endTime = \(endTime),progress = \(progress)")
            self?.playerView.pause()
            // 修改最新值
            self?.stuckPointMusicData?.startTime = Float64(startTime)
            self?.stuckPointMusicData?.endTime = Float64(endTime)
            // 红的指针完成
            if type == 3 {
                if CMTimeGetSeconds(self?.playerView.mPlayeTimeRange?.end ?? .zero) == 0 {
                    BFLog(message: "mPlayeTimeRange is error")
                    return
                }
                let newBeginSconds = (Double(startTime) + (Double(endTime) - Double(startTime)) * Double(progress)) * 600
                BFLog(message: " newBeginSconds is \(newBeginSconds)")
                let seekTimeRange: CMTimeRange = CMTimeRange(start: CMTime(value: CMTimeValue(Int64(newBeginSconds)), timescale: 600), end:
                    CMTime(value: CMTimeValue(Int64(endTime * 600)), timescale: 600))
                BFLog(message: "修改的开始 \(CMTimeGetSeconds(seekTimeRange.start)) 结束  \(CMTimeGetSeconds(seekTimeRange.end))")
                // 重新设置有效缓存
                self?.playerView.configCache(beginTime: CMTimeGetSeconds(seekTimeRange.start))
                self?.playerView.play(pauseFirstFrame: false, playeTimeRange: seekTimeRange)

            } else {
                // 更改素材开始时间及结束时间
                self?.projectModel.sData?.getBGMSession()?.sectionTimeline?.audioTrack?.audioTrackMaterials.first?.out = Float64(endTime)
                self?.projectModel.sData?.getBGMSession()?.sectionTimeline?.audioTrack?.audioTrackMaterials.first?.model_in = Float64(startTime)
                self?.projectModel.sData?.getBGMSession()?.sectionTimeline?.audioTrack?.audioTrackMaterials.first?.timelineIn = Float64(startTime)
                self?.projectModel.sData?.getBGMSession()?.sectionTimeline?.audioTrack?.audioTrackMaterials.first?.timelineOut = Float64(endTime)
                BFLog(message: "调整后总时长: \(endTime - startTime) startTime:\(startTime) endTime:\(endTime)")
                // 初始化音频的开始和结束时间
                self?.playeTimeRange = CMTimeRange(start: CMTimeMakeWithSeconds(Float64(startTime), preferredTimescale: BASE_FILTER_TIMESCALE), end: CMTimeMakeWithSeconds(Float64(endTime), preferredTimescale: BASE_FILTER_TIMESCALE))

                DispatchQueue.global().async { // 并行、异步
                    let beginTime: TimeInterval = Date().timeIntervalSince1970
                    self?.mStickers = self?.createStickers(sections: self?.projectModel.sData?.sections ?? List(), inputSize: CGSize(width: CGFloat(self?.projectModel.sData?.videoMetaData?.videoWidth ?? 0), height: CGFloat(self?.projectModel.sData?.videoMetaData?.videoHeight ?? 0)))

                    self?.playerView.mStickers = self?.mStickers
                    DispatchQueue.main.async { // 串行、异步
                        var endTime: TimeInterval = Date().timeIntervalSince1970
                        BFLog(message: "endTime is endTimeendTime \(endTime = beginTime)")
                        self?.playerView.play(pauseFirstFrame: false, playeTimeRange: self!.playeTimeRange)
                    }
                }
            }
            // 埋点上报
            if type == 1 || type == 2 {
                // 点击上报：拖动拖拽条（左/右部分）r
                PQEventTrackViewModel.baseReportUpload(businessType: .bt_buttonClick, objectType: type == 1 ? .ot_click_dragFront : .ot_click_dragBehind, pageSource: .sp_stuck_previewSyncedUp, extParams: ["targetTime": type == 1 ? startTime * 1000 : endTime * 1000], remindmsg: "点击上报：拖动拖拽条（左/右部分）")
            }
        }
        return stuckPointCuttingView
    }()

    /// 卡点模式标题
    lazy var pointEditRemindLab: UILabel = {
        let pointEditRemindLab = UILabel()
        pointEditRemindLab.backgroundColor = .clear
        pointEditRemindLab.textAlignment = .left
        pointEditRemindLab.font = UIFont.boldSystemFont(ofSize: 14)
        pointEditRemindLab.textColor = .black
        pointEditRemindLab.text = "卡点模式"
        return pointEditRemindLab
    }()

    /// 卡点模式标题
    lazy var speedTitleLab: UILabel = {
        let speedTitleLab = UILabel()
        speedTitleLab.backgroundColor = .clear
        speedTitleLab.textAlignment = .left
        speedTitleLab.font = UIFont.boldSystemFont(ofSize: 14)
        speedTitleLab.textColor = .black
        speedTitleLab.text = "节奏变化"
        return speedTitleLab
    }()

    /// 卡点模式下方操作区背景
    lazy var pointEditBGView: UIView = {
        let pointEditBGView = UIView()
        pointEditBGView.backgroundColor = .clear
        return pointEditBGView
    }()
    
    /// 下方音乐编辑操作区背景
    lazy var musicEditBGView: PQSelecteMusicView = {
        let musicEditBGView = PQSelecteMusicView.init()
        musicEditBGView.backgroundColor = .clear
        musicEditBGView.isUserInteractionEnabled = true
        musicEditBGView.isHidden = true
        musicEditBGView.musicSearchBtn.addTarget(self, action: #selector(musicSearchBtnClick(sender:)), for: .touchUpInside)
        
        musicEditBGView.didSelectItemHandle = { [weak self] isplaying in
            self?.playerView.pause()
        }
        musicEditBGView.btnClickHandle = { [weak self] _, bgmData in
            // 使用音乐
            self?.userstuckPointMusic(musicData: bgmData as? PQVoiceModel)

        }
        
        return musicEditBGView
    }()
 
    


    // 卡点编辑 btn
    lazy var pointEditerBtn: UIButton = {
        let pointEdterBtn = UIButton(type: .custom)
        pointEdterBtn.setImage(UIImage().BF_Image(named: "pointEditerBtn_n"), for: .normal)
        pointEdterBtn.setImage(UIImage().BF_Image(named: "pointEditerBtn_h"), for: .selected)
        pointEdterBtn.addTarget(self, action: #selector(pointEditerBtnClick(sender:)), for: .touchUpInside)
        pointEdterBtn.isSelected = true
        pointEdterBtn.adjustsImageWhenHighlighted = false
        return pointEdterBtn
    }()

    // 音乐编辑 btn
    lazy var musicEditerBtn: UIButton = {
        let musicEditerBtn = UIButton(type: .custom)
        musicEditerBtn.setImage(UIImage().BF_Image(named: "musicEditerBtn_n"), for: .normal)
        musicEditerBtn.setImage(UIImage().BF_Image(named: "musicEditerBtn_h"), for: .selected)
        musicEditerBtn.addTarget(self, action: #selector(musicEditerBtnClick(sender:)), for: .touchUpInside)
        musicEditerBtn.adjustsImageWhenHighlighted = false
        return musicEditerBtn
    }()

    // 快慢速卡点模式 btn
    lazy var speedStuckBtn: UIButton = {
        let speedStuckBtn = UIButton(type: .custom)
        speedStuckBtn.addTarget(self, action: #selector(editModelClick(sender:)), for: .touchUpInside)
        speedStuckBtn.setBackgroundImage(UIImage().BF_Image(named: "speedstuck_n"), for: .normal)
        speedStuckBtn.setBackgroundImage(UIImage().BF_Image(named: "speedstuck_h"), for: .selected)

        speedStuckBtn.tag = 1
        speedStuckBtn.adjustsImageWhenHighlighted = false
        return speedStuckBtn
    }()
    //
    lazy var speedStuckBtnGif: UIImageView = {
        let speedStuckBtnGif = UIImageView.init()
        speedStuckBtnGif.kf.setImage(with: URL(fileURLWithPath:  Bundle.init().BF_mainbundle().path(forResource: "speedstuck_h", ofType: ".gif")!))
        return speedStuckBtnGif

    }()

    // 跳转卡点模式 btn
    lazy var jumpPointBtn: UIButton = {
        let jumpPointBtn = UIButton(type: .custom)
        jumpPointBtn.setImage(UIImage().BF_Image(named: "jumpPoint_n"), for: .normal)
        jumpPointBtn.setImage(UIImage().BF_Image(named: "jumpPoint_h"), for: .selected)
        jumpPointBtn.tag = 2
        jumpPointBtn.addTarget(self, action: #selector(editModelClick(sender:)), for: .touchUpInside)
        jumpPointBtn.adjustsImageWhenHighlighted = false
        return jumpPointBtn
    }()
    lazy var jumpPointBtnGif: UIImageView = {
        let jumpPointBtnGif = UIImageView.init()
        jumpPointBtnGif.kf.setImage(with: URL(fileURLWithPath:  Bundle.init().BF_mainbundle().path(forResource: "jumpPoint_n", ofType: ".gif")!))
        return jumpPointBtnGif

    }()

    // 仅配乐模式 btn
    lazy var onlyMusicBtn: UIButton = {
        let onlyMusicBtn = UIButton(type: .custom)
        onlyMusicBtn.setImage(UIImage().BF_Image(named: "onlyMusic_n"), for: .normal)
        onlyMusicBtn.setImage(UIImage().BF_Image(named: "onlyMusic_h"), for: .selected)
        onlyMusicBtn.tag = 3
        onlyMusicBtn.addTarget(self, action: #selector(editModelClick(sender:)), for: .touchUpInside)
        onlyMusicBtn.adjustsImageWhenHighlighted = false
        return onlyMusicBtn
    }()

    // 编辑模式的指示线
    lazy var editModelLineView: UIView = {
        let editModelLineView = UIView()
        editModelLineView.backgroundColor = UIColor.hexColor(hexadecimal: "#EFEFEF")
        return editModelLineView
    }()

    // 操作面板上的分割线
    lazy var optionlineView: UIView = {
        let optionlineView = UIView()
        optionlineView.backgroundColor = UIColor.hexColor(hexadecimal: "#EFEFEF")
        return optionlineView
    }()

    // 固定速度 UI
    lazy var speedSettingView: PQSpeedSettingView = {
        let speedSettingView = PQSpeedSettingView()
        speedSettingView.backgroundColor = .clear
        speedSettingView.viewType = 1
        speedSettingView.selectSpeedCallBack = { [weak self] maxSpeed, minSpeed,selectIndex in
            BFLog(message: "固定maxSpeed is\(maxSpeed) minSpeed \(minSpeed)")
            //更新最后一次选择的位置恢复时使用
            if(speedSettingView.viewType == 2){
                self?.lastJumpSpeedSelectIndex = selectIndex
            }else{
                self?.lastSpeedSelectIndex = selectIndex
            }
            self?.maxSpeed = maxSpeed
            self?.minSpeed = minSpeed
            if maxSpeed == 0 && minSpeed == 0 {
                self?.customSpeedSettingView.isHidden = false
                self?.customSpeedSettingView.isJumpSpeedModel = speedSettingView.viewType == 2
            }
        }
        return speedSettingView

    }()

    // 自定义速度
    lazy var customSpeedSettingView: PQCustomSpeedSettingView = {
        let customSpeedSettingView = PQCustomSpeedSettingView(frame: CGRect(x: 0, y: cScreenHeigth - 354, width: cScreenWidth, height: 354))
        customSpeedSettingView.isHidden = true
        customSpeedSettingView.isJumpSpeedModel = false
        customSpeedSettingView.selectSpeedCallBack = { [weak self] maxSpeed, minSpeed ,isJumpSpeedModel in
            self?.maxSpeed = maxSpeed
            self?.minSpeed = minSpeed
            BFLog(message: "自定义速度maxSpeed is\(maxSpeed) minSpeed \(minSpeed) \(isJumpSpeedModel)")
        
        }
        return customSpeedSettingView

    }()

    /// 音乐标题
    lazy var musicNameView: UIView = {
        let musicNameView = UIView()
        musicNameView.addSubview(musicNameLab)
        let nameWidth: CGFloat = musicNameLab.frame.width + (25 + cDefaultMargin * 3)
        musicNameView.frame = CGRect(x: (view.frame.width - nameWidth) / 2, y: cDevice_iPhoneStatusBarHei + (cDevice_iPhoneNavBarHei - cDefaultMargin * 3) / 2, width: nameWidth, height: cDefaultMargin * 3)
//        musicNameView.backgroundColor = UIColor.hexColor(hexadecimal: "#333333")
        musicNameView.addCorner(corner: musicNameView.frame.height / 2)
        let musicImageView = UIImageView()
        musicImageView.tintColor = PQBFConfig.shared.styleTitleColor
        musicImageView.image = UIImage().BF_Image(named: "stuckPoint_reCreate_music").withRenderingMode(.alwaysTemplate)
        musicImageView.frame = CGRect(x: musicNameView.frame.height / 2 - 5, y: (musicNameView.frame.height - 22) / 2, width: 22, height: 22)
        musicNameView.addSubview(musicImageView)
        musicNameLab.frame.origin.x = musicImageView.frame.maxX + 5
        return musicNameView
    }()

    /// 音乐歌曲名称
    lazy var musicNameLab: LMJHorizontalScrollText = {
        let nameWidth: CGFloat = sizeWithText(text: "\(stuckPointMusicData?.musicName ?? "")", font: UIFont.systemFont(ofSize: 13), size: CGSize(width: view.frame.width - ((cDefaultMargin * 6 + 16 * 2) * 2) - (25 + cDefaultMargin * 3), height: cDefaultMargin * 3)).width
        let musicNameLab = LMJHorizontalScrollText(frame: CGRect(x: 0, y: 0, width: nameWidth < cDefaultMargin * 4 ? cDefaultMargin * 4 : nameWidth, height: cDefaultMargin * 3))
        musicNameLab.textColor = PQBFConfig.shared.styleTitleColor
        musicNameLab.textFont = UIFont.systemFont(ofSize: 13)
        musicNameLab.speed = 0.03
        musicNameLab.moveDirection = LMJTextScrollMoveLeft
        musicNameLab.moveMode = LMJTextScrollContinuous
        if nameWidth < cDefaultMargin * 4 {
            musicNameLab.text = " \(stuckPointMusicData?.musicName ?? "") "
        } else {
            musicNameLab.text = " \(stuckPointMusicData?.musicName ?? "") "
        }
        return musicNameLab
    }()

    /// 同步进度显示
    lazy var synchroMarskView: PQStuckPointLoadingView = {
        var synchroMarskView = PQStuckPointLoadingView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
        synchroMarskView.cancelHandle = { [weak self] _ in
            self?.navigationController?.popViewController(animated: true)
        }
        return synchroMarskView
    }()

    override func viewWillAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        lineView?.isHidden = true
        UIApplication.shared.isIdleTimerDisabled = true
        musicNameLab.move()

        PQNotification.addObserver(self, selector: #selector(enterBackground), name: UIApplication.didEnterBackgroundNotification, object: nil)
    }

    @objc func enterBackground() {
        BFLog(message: "进入到后台")
        // 取消导出
        playerView.pause()
    }

    override func backBtnClick() {
        super.backBtnClick()
        playerView.pause()
        // 点击上报：返回按钮
        PQEventTrackViewModel.baseReportUpload(businessType: .bt_buttonClick, objectType: .ot_click_back, pageSource: .sp_stuck_previewSyncedUp, extParams: nil, remindmsg: "卡点视频数据上报-（点击上报：返回按钮）")
    }
    
    //使用选择音乐 调用情况：1，操作面板直接选择 ，2 搜索界面点击使用
    func userstuckPointMusic(musicData: PQVoiceModel?) {
        
        //1，音乐面板点击了使用
        stuckPointMusicData = musicData
        //2，同步最新音乐数据
        synchroMusicInfoData()
        
        //3，更新音乐标题UI
        let nameWidth: CGFloat = sizeWithText(text: "\(stuckPointMusicData?.musicName ?? "")", font: UIFont.systemFont(ofSize: 13), size: CGSize(width: (view.frame.width ) - ((cDefaultMargin * 6 + 16 * 2) * 2) - (25 + cDefaultMargin * 3), height: cDefaultMargin * 3)).width
        
        if nameWidth < cDefaultMargin * 4 {
            musicNameLab.text = " \(stuckPointMusicData?.musicName ?? "") "
        } else {
            musicNameLab.text = " \(stuckPointMusicData?.musicName ?? "") "
        }
    }
    //点击搜索音乐
    @objc func musicSearchBtnClick(sender: UIButton) {
        let searchVC = PQEditMusicSearchController.init()
        
        searchVC.btnClickHandle = { [weak self] _, bgmData in
            // 使用音乐
            self?.userstuckPointMusic(musicData: bgmData as? PQVoiceModel)
        }
        
        let navigationController: UINavigationController = UINavigationController(rootViewController: searchVC)
        navigationController.modalPresentationStyle = .fullScreen
        present(navigationController, animated: true, completion: nil)
    }
    
    // 卡点编辑
    @objc func pointEditerBtnClick(sender: UIButton) {
        if sender.isSelected { return }
        sender.isSelected = !sender.isSelected
        musicEditerBtn.isSelected = false
        pointEditBGView.isHidden = false
        musicEditBGView.isHidden = true
    }

    // 音乐编辑
    @objc func musicEditerBtnClick(sender: UIButton) {
        if sender.isSelected { return }
        sender.isSelected = !sender.isSelected
        pointEditerBtn.isSelected = false
        pointEditBGView.isHidden = true
        musicEditBGView.isHidden = false
    }

    // 三种模式修改
    @objc func editModelClick(sender: UIButton) {
        if sender.isSelected {
            BFLog(message: "已经是选中状态")
            return
        }
        lastEditModelBtn?.isSelected = false
        sender.isSelected = !sender.isSelected
        lastEditModelBtn = sender

        BFLog(message: "sender tag is \(sender.tag)")
        // 1 ui 调整
        if sender.tag == 1 || sender.tag == 2 {
            speedSettingView.viewType = sender.tag
            customSpeedSettingView.isJumpSpeedModel = speedSettingView.viewType == 2

            speedSettingView.snp.remakeConstraints { make in
                make.left.equalToSuperview().offset(16)
                make.right.equalToSuperview()
                make.top.equalTo(editModelLineView.snp_bottom).offset(8)
                make.height.equalTo(sender.tag == 1 ? 44 : 30)
            }
            speedSettingView.isHidden = false
            speedTitleLab.isHidden = false
            sustomSwitchView.isHidden = false
            editModelLineView.isHidden = false
        } else {
            speedSettingView.isHidden = true
            speedTitleLab.isHidden = true
            sustomSwitchView.isHidden = true
            editModelLineView.isHidden = true
        }
        // 2 不同状的业务逻辑
        var speedStuckBtnGifName = ""
        var jumpPointBtnGifName = ""
        if sender.tag == 1 { // 快慢速
            speedStuckBtnGifName = "speedstuck_h"
            jumpPointBtnGifName = "jumpPoint_n"
            currentCreateStickersModel = .createStickersModelSpeed
            speedSettingView.setSelectItem(index: lastSpeedSelectIndex)
         
            
        } else if sender.tag == 2 { // 跳跃卡点
            speedStuckBtnGifName = "speedstuck_n"
            jumpPointBtnGifName = "jumpPoint_h"
            speedSettingView.setSelectItem(index: lastJumpSpeedSelectIndex)
            currentCreateStickersModel = .createStickersModelPoint

        } else if sender.tag == 3 { // 仅音乐
            speedStuckBtnGifName = "speedstuck_n"
            jumpPointBtnGifName = "jumpPoint_n"
            currentCreateStickersModel = .createStickersModelOnlyMusic

        }
        
        settingPlayerView()
        speedStuckBtnGif.kf.setImage(with: URL(fileURLWithPath:  Bundle.init().BF_mainbundle().path(forResource:speedStuckBtnGifName , ofType: ".gif")!))
        jumpPointBtnGif.kf.setImage(with: URL(fileURLWithPath:  Bundle.init().BF_mainbundle().path(forResource: jumpPointBtnGifName  , ofType: ".gif")!))
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        UIApplication.shared.isIdleTimerDisabled = false
        musicNameLab.stop()
        playerView.pause()
        
        musicEditBGView.pausePlayer()
    }

    override func viewDidLoad() {
        super.viewDidLoad()
        leftButton(image: "icon_detail_back", tintColor: PQBFConfig.shared.styleTitleColor)
        navHeadImageView?.addSubview(nextBtn)
        navHeadImageView?.addSubview(musicNameView)

        // 添加子视图
        addSubViews()
        // 导出相册视频
        exportPhotoData()
        // 同步音乐数据
        synchroMusicInfoData()
        // 曝光上报：预览页面曝光上报
        PQEventTrackViewModel.baseReportUpload(businessType: .bt_windowView, objectType: .ot_view_previewSyncedUp, pageSource: .sp_stuck_previewSyncedUp, extParams: nil, remindmsg: "卡点视频数据上报-（曝光上报：预览页面曝光上报）")
    }

    /// 添加子视图
    /// - Returns: <#description#>
    func addSubViews() {
        if (stuckPointMusicData?.rhythmSdata.count ?? 0) <= 0 {
            return
        }
        view.addSubview(playerView)
        view.addSubview(pointEditBGView)
        view.addSubview(musicEditBGView)
        view.addSubview(optionlineView)

        view.addSubview(pointEditerBtn)
        view.addSubview(musicEditerBtn)
        view.addSubview(customSpeedSettingView)

        //卡点
        pointEditBGView.addSubview(pointEditRemindLab)
        pointEditBGView.addSubview(speedTitleLab)
        pointEditBGView.addSubview(speedStuckBtn)
        speedStuckBtn.addSubview(speedStuckBtnGif)
        pointEditBGView.addSubview(jumpPointBtn)
        jumpPointBtn.addSubview(jumpPointBtnGif)
        pointEditBGView.addSubview(onlyMusicBtn)
        pointEditBGView.addSubview(editModelLineView)
        pointEditBGView.addSubview(speedSettingView)
        pointEditBGView.addSubview(sustomSwitchView)
        
        //音乐
        musicEditBGView.addSubview(stuckPointCuttingView)


        pointEditerBtn.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(100)
            make.bottom.equalToSuperview().offset(-(14 + cAKSafeAreaHeight))
            make.height.equalTo(24)
            make.width.equalTo(24)
        }

        musicEditerBtn.snp.makeConstraints { make in
            make.right.equalToSuperview().offset(-100)
            make.top.equalTo(pointEditerBtn.snp_top)
            make.height.equalTo(24)
            make.width.equalTo(24)
        }
        pointEditBGView.snp.makeConstraints { make in
            make.left.right.equalTo(view)
            make.bottom.equalTo(pointEditerBtn.snp_top).offset(-7)
            make.height.equalTo(290)
        }
        musicEditBGView.snp.makeConstraints { make in
            make.left.right.equalTo(view)
            make.bottom.equalTo(pointEditerBtn.snp_top).offset(-7)
            make.height.equalTo(290)
        }
        stuckPointCuttingView.snp.makeConstraints { make in
            make.left.right.equalTo(view)
            make.bottom.equalTo(pointEditerBtn.snp_top).offset(-7)
            make.height.equalTo(290)
        }
        optionlineView.snp.makeConstraints { make in
            make.left.right.equalToSuperview()
            make.bottom.equalTo(pointEditBGView.snp_bottom).offset(-1)
            make.height.equalTo(1)
        }
        
        stuckPointCuttingView.snp.makeConstraints { make in
            make.left.right.equalToSuperview()
            make.bottom.equalTo(musicEditBGView.snp_bottom).offset(-1)
            make.height.equalTo(85)
        }
        pointEditRemindLab.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.top.equalToSuperview().offset(16)
            make.height.equalTo(20)
            make.width.equalTo(80)
        }
        speedStuckBtn.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.top.equalTo(pointEditRemindLab.snp_bottom).offset(8)
            make.height.equalTo(80)
            make.width.equalTo(80)
        }
        speedStuckBtnGif.snp.makeConstraints { make in
            make.height.width.equalTo(46)
            make.top.equalToSuperview().offset(8)
            make.centerX.equalToSuperview()
        }

        jumpPointBtn.snp.makeConstraints { make in
            make.left.equalTo(speedStuckBtn.snp_right).offset(10)
            make.top.equalTo(speedStuckBtn.snp_top)
            make.height.equalTo(80)
            make.width.equalTo(80)
        }
        jumpPointBtnGif.snp.makeConstraints { make in
            make.height.width.equalTo(46)
            make.top.equalToSuperview().offset(8)
            make.centerX.equalToSuperview()
        }

        onlyMusicBtn.snp.makeConstraints { make in
            make.left.equalTo(jumpPointBtn.snp_right).offset(10)
            make.top.equalTo(speedStuckBtn.snp_top)
            make.height.equalTo(80)
            make.width.equalTo(64)
        }
        editModelLineView.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.right.equalToSuperview()
            make.bottom.equalTo(onlyMusicBtn.snp_bottom).offset(17)
            make.height.equalTo(1)
        }
        speedSettingView.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.right.equalToSuperview()
            make.top.equalTo(editModelLineView.snp_bottom).offset(8)
            make.height.equalTo(44)
        }
        speedTitleLab.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.top.equalToSuperview().offset(213)
            make.height.equalTo(20)
            make.width.equalTo(80)
        }
        sustomSwitchView.snp.makeConstraints { make in
            make.left.equalToSuperview().offset(16)
            make.top.equalTo(speedTitleLab.snp_bottom).offset(8)
            make.height.equalTo(30)
            make.width.equalTo(180)
        }
    }

    @objc func nextBtnClick(sender _: UIButton) {
        BFLog(message: "去发布")
        playerView.pause()
        let videoExporter = PQStuckPointPublicController()
        videoExporter.isReCreate = isReCreate
        videoExporter.selectedTotalDuration = selectedTotalDuration
        videoExporter.selectedDataCount = selectedDataCount
        videoExporter.selectedImageDataCount = selectedImageDataCount
        // 使用深 copy
        let json = projectModel.toJSONString(prettyPrint: false)
        if json == nil {
            BFLog(message: "数据转换有问题 跳转")
            return
        }
        let tempModel: PQEditProjectModel? = Mapper<PQEditProjectModel>().map(JSONString: json!)
        videoExporter.mStickers = mStickers
        videoExporter.audioMixModel = stuckPointMusicData
        videoExporter.editProjectModel = tempModel
        navigationController?.pushViewController(videoExporter, animated: true)
        // 点击上报：去合成
        PQEventTrackViewModel.baseReportUpload(businessType: .bt_buttonClick, objectType: .ot_click_commit, pageSource: .sp_stuck_previewSyncedUp, extParams: ["musicName": stuckPointMusicData?.musicName ?? "", "musicId": stuckPointMusicData?.musicId ?? "", "rhythmNumber": stuckPointMusicData?.speed ?? 2, "duration": ((stuckPointMusicData?.endTime ?? 0) - (stuckPointMusicData?.startTime ?? 0)) * 1000], remindmsg: "点击上报：去合成")
    }

    // MARK: - 播放器相关操作

    /// seek 播放器
    /// - Parameter playeTimeRange: 开始和结束时间
    func seekPlayer(playeTimeRange: CMTimeRange) {
        playerView.setEnableSeek(isSeek: true)
        playerView.play(pauseFirstFrame: false, playeTimeRange: playeTimeRange)
    }

    /// 通过传入的    selectedPhotoData 、 stuckPointMusicData 创建 projectModel 模型 后面都使用 projectModel 参数
    func createPorjectData() {
        // 1,添加选择的视觉素材
        let section: PQEditSectionModel = PQEditSectionModel()
        selectedPhotoData?.forEach { model in

            let json = model.toJSONString(prettyPrint: false)
            if json == nil {
                BFLog(message: "数据转换有问题 跳转")
                return
            }
            let tempModel: PQEditVisionTrackMaterialsModel = Mapper<PQEditVisionTrackMaterialsModel>().map(JSONString: json!)!
            section.sectionTimeline?.visionTrack?.visionTrackMaterials.append(tempModel)
        }

        projectModel.sData?.sections.append(section)

        // 2,添加背景音乐
        projectModel.sData?.addBGM(audioMix: stuckPointMusicData!)
    }

    // 设置播放器
    func settingPlayerView() {
        // 1,设置播放器的显示区域 和画布大小
        //  - 按第一个素材尺寸自适应
        let playerShowHeight = pointEditBGView.frame.minY - (navHeadImageView?.frame.maxY ?? 0)
        var showRect: CGRect = PQPlayerViewModel.getShowCanvasRect(editProjectModel: projectModel, showType: 1, playerViewHeight: playerShowHeight)
        if showRect.size.width == showRect.size.height {
            if cScreenWidth > playerShowHeight {
                showRect.origin.x = (cScreenWidth - playerShowHeight) / 2
                showRect.size.width = playerShowHeight
                showRect.size.height = playerShowHeight
            } else {
                showRect.origin.x = 0
                showRect.size.width = cScreenWidth
                showRect.size.height = cScreenWidth
            }
        }
        showRect.origin.y = (playerShowHeight - showRect.size.height) / 2.0 + (navHeadImageView?.frame.maxY ?? 0)
        if showRect.size.width != 0, showRect.size.height != 0 {
            playerView.resetCanvasFrame(frame: showRect)
        }

        var firstModel: PQEditVisionTrackMaterialsModel?
        for part in projectModel.sData!.sections {
            if part.sectionTimeline?.visionTrack?.getEnableVisionTrackMaterials().count ?? 0 > 0 {
                firstModel = part.sectionTimeline?.visionTrack?.getEnableVisionTrackMaterials().first
                break
            }
        }

        var videoSize: CGSize = CGSize(width: Int(firstModel?.width ?? 0), height: Int(firstModel?.height ?? 0))

        var minSlider = min(videoSize.width, videoSize.height)
        var maxSlider = max(videoSize.width, videoSize.height)
        let ration = 1080 / minSlider
        minSlider = minSlider * ration
        maxSlider = maxSlider * ration
        if videoSize.width > videoSize.height { // 宽屏
            videoSize = CGSize(width: maxSlider, height: minSlider)
        } else {
            videoSize = CGSize(width: minSlider, height: maxSlider)
        }

        let maxValue = max(videoSize.width, videoSize.height )
        if maxValue > 1920 {
            let maxRation = 1920 / maxValue

            videoSize = CGSize(width: videoSize.width * CGFloat(maxRation), height: videoSize.height * CGFloat(maxRation))

            BFLog(message: "最长边已经超过 1920 要等比缩小 缩放后\(videoSize)")
        }

        if (Int(videoSize.width) % 2) != 0 {
            videoSize.width = videoSize.width - 1
        }
        if (Int(videoSize.height) % 2) != 0 {
            videoSize.height = videoSize.height - 1
        }

        projectModel.sData?.videoMetaData?.videoWidth = Int(videoSize.width)
        projectModel.sData?.videoMetaData?.videoHeight = Int(videoSize.height)

        // 2,创建滤镜
        let beginTime: TimeInterval = Date().timeIntervalSince1970
        mStickers = createStickers(sections: projectModel.sData?.sections ?? List(), inputSize: CGSize(width: CGFloat(projectModel.sData?.videoMetaData?.videoWidth ?? 0), height: CGFloat(projectModel.sData?.videoMetaData?.videoHeight ?? 0)), model: currentCreateStickersModel)
        playerView.mStickers = mStickers

        let end: TimeInterval = Date().timeIntervalSince1970
        BFLog(message: "createStickers tiskskskskme  \(end - beginTime)")

        // 3,设置音频
        let audioPath = stuckPointMusicData?.localPath ?? ""
    
        BFLog(message: "初始化音频播放器的音频地址为:\(audioPath)")
        playerView.stop()
        // 这里的测试这个音乐播放有问题
//        self.playerView.updateAsset(URL(fileURLWithPath: "63930549652d74e477141e3b79c8d29a9ef8af81625053214516.mp3", relativeTo:Bundle.main.resourceURL!), videoComposition: nil, audioMixModel: nil)
        
        
        //设置音乐的拼接范围，开始：推荐的卡点  结束：点到的倒数第二位
        let lastSecondPoint =   Float((stuckPointMusicData!.rhythmSdata[0].pointTimes[stuckPointMusicData!.rhythmSdata[0].pointTimes.count - 2] )) / Float(BASE_FILTER_TIMESCALE)
        let clipAudioRange =
        CMTimeRange(start: CMTime(value: CMTimeValue( Float(stuckPointMusicData?.startTime ?? 0) * Float(BASE_FILTER_TIMESCALE)), timescale: BASE_FILTER_TIMESCALE), end: CMTime(value: CMTimeValue((Float(lastSecondPoint)) * Float(BASE_FILTER_TIMESCALE)), timescale: BASE_FILTER_TIMESCALE))
 
        playerView.updateAsset(URL(fileURLWithPath: documensDirectory + audioPath), videoComposition: nil, audioMixModel: nil,originMusicDuration:finallyUserAudioTime,clipAudioRange: clipAudioRange)
 
        let end2: TimeInterval = Date().timeIntervalSince1970
        BFLog(message: "updateAsset tiskskskskme  \(end2 - end)")
        // 4, 设置播放器的输出画布大小
        playerView.movie?.mShowVidoSize = CGSize(width: CGFloat(projectModel.sData?.videoMetaData?.videoWidth ?? 0), height: CGFloat(projectModel.sData?.videoMetaData?.videoHeight ?? 0))

        // 5,开始播放
        playerView.isLoop = false
        playerView.showProgressLab = true

        // 初始化音频的开始和结束时间
        BFLog(message: "播放的器 开始\(String(describing: CMTimeGetSeconds(playeTimeRange.start))) 结束 \(String(describing: CMTimeGetSeconds(playeTimeRange.end)))")
        let end3: TimeInterval = Date().timeIntervalSince1970

        playerView.play(pauseFirstFrame: false, playeTimeRange: CMTimeRange(start: playeTimeRange.start, end: playeTimeRange.end))

        let end4: TimeInterval = Date().timeIntervalSince1970
        BFLog(message: " playerView.play tiskskskskme  \(end4 - end3)")

        // 6,进度回调
        playerView.progress = { [weak self] currentTime, tatolTime, _ in

            // 更新进度
            let progress = (currentTime - CMTimeGetSeconds(self?.playeTimeRange.start ?? .zero)) / CMTimeGetSeconds(self?.playeTimeRange.duration ?? .zero)
            BFLog(message: "\(currentTime) \(tatolTime) 显示播放器进度为: \(progress)")

            self?.stuckPointCuttingView.updateProgress(progress: CGFloat(progress))

            if self?.synchroMarskView.superview != nil {
                self?.synchroMarskView.removeMarskView()
            }
        }
    }

    deinit {
        BFLog(message: "卡点视频预览界面销毁")
        musicNameLab.stop()
        playerView.pause()
        // 取消所有的导出
        PQSingletoMemoryUtil.shared.allExportSession.forEach { _, exportSession in
            exportSession.cancelExport()
        }
    }
}

// MARK: - 视频渲染相关逻辑方法

extension PQStuckPointEditerController {
    /// 分割视频 这里只设置视频类型的 in 和 out 并不设置显示的开始和结束时间   mp4 ,png ,png ,mp4
    /// - Parameter section: 当前段
    /// - Parameter stuckPoints: 用户选择的，或推荐的卡点数
    /// - Returns: 返回分割后的所有 stickers 和卡点数是一致的
    func clipVideoMerage(section: PQEditSectionModel, stuckPoints: [Float]) -> [PQEditVisionTrackMaterialsModel] {
        var stickers: Array = Array<PQEditVisionTrackMaterialsModel>.init()
        // 第二种情况：有视频要进行分割
        /*
         1, 确定每个视频素材需要切的段数p
         2, 将所有视频时长相加，得到总视频素材时长L = l1 + l2 + ... + ln
         3, 视频素材a1需要切分的个数clipNum = max (round (kongduan * a1 / L) , 1)
         */
        // 要补的空位数
        let kongduan: Int = Int(stuckPoints.count) - Int(section.sectionTimeline!.visionTrack?.getEnableVisionTrackMaterials().count ?? 0)

        // 所有视频总时长
        var videoTotalDuration: Float64 = 0.0
        for video in section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials(type: "video") {
            let asset: AVURLAsset = AVURLAsset(url: URL(fileURLWithPath: documensDirectory + video.locationPath), options: nil)
            videoTotalDuration = videoTotalDuration + Float64(CMTimeGetSeconds(asset.duration))
        }
        if videoTotalDuration == 0 {
            BFLog(message: "视频总时长出现错误！！！！这里应该有视频素材的")
            return stickers
        }
        for sticker in section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials() {
            if sticker.type == StickerType.VIDEO.rawValue {
                let asset: AVURLAsset = AVURLAsset(url: URL(fileURLWithPath: documensDirectory + sticker.locationPath), options: nil)
                // 要分割的段落
                let clipNum = Int(max(round(Double(kongduan) * CMTimeGetSeconds(asset.duration) / videoTotalDuration), 1))
                sticker.duration = CMTimeGetSeconds(asset.duration)
                BFLog(message: "单个视频\(sticker.locationPath)时长：:\(CMTimeGetSeconds(asset.duration)) ,clipNum is:\(clipNum)")
                for clipindex in 0 ... clipNum - 1 {
                    // deep copy sticker model 防止只有一个对象
                    let stickerjson = sticker.toJSONString(prettyPrint: false)
                    let deepCopySticker = Mapper<PQEditVisionTrackMaterialsModel>().map(JSONString: stickerjson!)

                    // 设置循环模式和适配模式
                    deepCopySticker?.generateDefaultValues()

                    deepCopySticker?.model_in = clipindex == 0 ? 0 : CMTimeGetSeconds(asset.duration) / Double(clipNum) * Double(clipindex)
                    deepCopySticker?.out = (deepCopySticker?.model_in ?? 0) + CMTimeGetSeconds(asset.duration) / Double(clipNum)

                    if (deepCopySticker?.model_in ?? 0) >= CMTimeGetSeconds(asset.duration) || (deepCopySticker?.out ?? 0) >= CMTimeGetSeconds(asset.duration) {
                        deepCopySticker?.model_in = CMTimeGetSeconds(asset.duration) - CMTimeGetSeconds(asset.duration) / Double(clipNum)

                        deepCopySticker?.out = CMTimeGetSeconds(asset.duration)
                    }

                    BFLog(message: " crilp is in \(deepCopySticker?.model_in ?? 0) out \(deepCopySticker?.out ?? 0) 总时长\(CMTimeGetSeconds(asset.duration))")

                    if deepCopySticker != nil {
                        stickers.append(deepCopySticker!)
                    }
                }
            } else if sticker.type == StickerType.IMAGE.rawValue {
                sticker.generateDefaultValues()
                stickers.append(sticker)
            }
        }

        return stickers
    }

    /// 创建sticker
    /// - Parameters:
    ///   - sections: 项目所有段落数据信息
    ///   - inputSize: 画布大小
    /// - Returns: filters 数据 播放器可直接使用
    func createStickers(sections: List<PQEditSectionModel>, inputSize _: CGSize = .zero, model: createStickersModel = .createStickersModelPoint) -> [PQEditVisionTrackMaterialsModel] {
        
        // 推荐卡点数
        dealParameter(model: model)
        let stuckPoints: Array = finallyStuckPoints
        
        // 保存滤镜对象数据
        var stickers: Array = Array<PQEditVisionTrackMaterialsModel>.init()
        if model == .createStickersModelPoint { //跳跃卡点
            for section in sections {
                if section.sectionType == "normal" {
                    // 推荐卡点数
                    var stuckPoints: Array = Array<Float>.init()

                    var stuckPointsTemp = Array<Float>.init()
                    for (index, dunshu) in stuckPointMusicData!.rhythmSdata[0].pointTimes.enumerated() {
                        BFLog(message: "原所有卡点数:\(index) \(Float(dunshu) / Float(BASE_FILTER_TIMESCALE))")
                        if Float64(dunshu) / Float64(BASE_FILTER_TIMESCALE) > CMTimeGetSeconds(playeTimeRange.start), Float64(dunshu) / Float64(BASE_FILTER_TIMESCALE) < CMTimeGetSeconds(playeTimeRange.end) {
                            stuckPointsTemp.append(Float(dunshu) / Float(BASE_FILTER_TIMESCALE))
                        }
                    }

                    /* 快慢速模式下取卡点 1 3 5
                     - 快节奏为选中区域的所有点位，即0，1，2，3，4 5 6 7 8 9 10 ……
                     - 适中为每两个点位取一个，即0，3，6，9 12
                     - 慢节奏为每三个点位取一个，即0 5 10 15
                     慢节奏要做特殊处理
                 
                     */
                    // 跳跃卡点模式下根据不同速度 取卡点 1,2,3
                    /*
                     - 快节奏为选中区域的所有点位，即0，1，2，3，4……
                     - 适中为每两个点位取一个，即0，2，4，6……
                     - 慢节奏为每三个点位取一个，即0，3，6，9……
                     */
                    BFLog(message: "stuckPointMusicData?.speed is \(String(describing: stuckPointMusicData?.speed))")
                    for (index, point) in stuckPointsTemp.enumerated() {
                        if stuckPointMusicData?.speed == 1 {
                            stuckPoints.append(Float(point))
                        } else if stuckPointMusicData?.speed == 2 {
                            if index % 2 == 0 {
                                stuckPoints.append(point)
                            }

                        } else if stuckPointMusicData?.speed == 3 {
                            if index % 3 == 0 {
                                stuckPoints.append(point)
                            }
                        }
                    }

                    for point in stuckPoints {
                        BFLog(message: "没有 start end 计算后的卡点数\(point)")
                    }
                    if stuckPoints.first != nil {
                        stuckPoints.removeFirst()
                    }
                    if stuckPoints.last != nil {
                        stuckPoints.removeLast()
                    }
                    // 开始时间是服务器返回， 结果时间根据策略计算的
                    stuckPoints.insert(Float(CMTimeGetSeconds(playeTimeRange.start)), at: 0)
                    stuckPoints.insert((Float(CMTimeGetSeconds(playeTimeRange.end))), at: stuckPoints.count)

                    for point in stuckPoints {
                        BFLog(message: "有 start end 计算后的卡点数\(point)")
                    }

                    BFLog(message: "stuckPoints count is \(stuckPoints.count)")
                    // 当用户上传视觉素材个数大于等于音乐选择区域节拍分割个数时，无需进行视频分割，只显示卡点数-1 个素材
                    if section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials().count >= stuckPointMusicData!.rhythmSdata[0].pointTimes.count {
                        for (index, sticker) in section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials().enumerated() {
                            if index == stuckPointMusicData!.rhythmSdata[0].pointTimes.count {
                                BFLog(message: "到达卡点数量")
                                break
                            }
                            BFLog(message: "创建 filter start :\(sticker.timelineIn) end :\(sticker.timelineOut) type is \(sticker.type) \(sticker.locationPath)")
                            sticker.timelineIn = Float64("\(stuckPoints[index])") ?? 0.0
                            sticker.timelineOut = Float64("\(stuckPoints[index + 1])") ?? 0.0
                            BFLog(message: "卡点 间隔 \(sticker.timelineIn - sticker.timelineOut)")
                            sticker.generateDefaultValues()
                            stickers.append(sticker)
                        }

                    } else {
                        // 卡点数 > 选择素材数

                        // 第一种情况：全是图片，图片回环播放
                        if section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials(type: "video").count == 0, section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials(type: "image").count > 0 {
                            for (index, point) in stuckPoints.enumerated() {
                                let sticker: PQEditVisionTrackMaterialsModel = section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials()[index % section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials().count]
                                BFLog(message: "stickerlocationPath sticker : \(sticker.locationPath)")
                                let stickerjson = sticker.toJSONString(prettyPrint: false)

                                let deepCopySticker = Mapper<PQEditVisionTrackMaterialsModel>().map(JSONString: stickerjson!)
                                if deepCopySticker!.type == StickerType.IMAGE.rawValue {
                                    if index + 1 < stuckPoints.count {
                                        deepCopySticker!.timelineIn = Float64("\(stuckPoints[index])") ?? 0.0
                                        deepCopySticker!.timelineOut = Float64("\(stuckPoints[index + 1])") ?? 0.0
                                        if deepCopySticker != nil {
                                            deepCopySticker?.generateDefaultValues()
                                            stickers.append(deepCopySticker!)
                                        }
                                    }
                                }
                            }
                        } else {
                            // 第二种情况：有视频要进行分割
                            let clipFilters = clipVideoMerage(section: section, stuckPoints: stuckPoints)
                            for (index, point) in stuckPoints.enumerated() {
                                BFLog(message: "aaaaaindexindeindexxindexindexindex \(index) \(point)")
                                if index + 1 < stuckPoints.count, index < clipFilters.count {
                                    BFLog(message: "bbbbbindexindeindexxindexindexindex \(index) \(point)")
                                    let sticker: PQEditVisionTrackMaterialsModel = clipFilters[index]
                                    sticker.timelineIn = Float64("\(stuckPoints[index])") ?? 0.0
                                 
                                    sticker.timelineOut = Float64("\(stuckPoints[index + 1])") ?? 0.0
                                    // 卡点的时间 >  in out 值 这里就会出现鬼畜效果
                                    let timelineInterval = sticker.timelineOut - sticker.timelineIn
                                    let inOutInterval = sticker.out - sticker.model_in
                                    if timelineInterval > inOutInterval {
                                        BFLog(message: "实际要显示卡点时长\(timelineInterval) 素材裁剪时长：\(inOutInterval)")
                                        sticker.out = sticker.model_in + timelineInterval

                                        // 下面只是 LOG 方便查问题
                                        let stickerInOut = sticker.out - sticker.model_in
                                        let stickerTimelineInOut = sticker.timelineOut - sticker.timelineIn
                                        if stickerInOut != stickerTimelineInOut {
                                            BFLog(message: "sticker.timelineIn \(sticker.timelineIn) stickerTimelineInOut is\(stickerTimelineInOut) stickerInOut is\(stickerInOut) 相差\(stickerTimelineInOut - stickerInOut)")
                                        }
                                    }

                                    // out > 素材的总时长in out 进行前移操作
                                    let offsetAssetDuration = sticker.out - sticker.duration
                                    if offsetAssetDuration > 0 {
                                        sticker.model_in = sticker.model_in - offsetAssetDuration
                                        sticker.out = sticker.out - offsetAssetDuration
                                    }

                                    BFLog(message: "index is \(index)分割后 创建 filter timelineIn :\(sticker.timelineIn) timelineOut :\(sticker.timelineOut)  in :\(sticker.model_in) out:\(sticker.out) type is \(sticker.type) 显示总时长为：\(sticker.timelineOut - sticker.timelineIn)  裁剪总时长\(sticker.out - sticker.model_in)")

                                    stickers.append(sticker)
                                }
                            }
                        }
                    }
                }
            }
        } else if model == .createStickersModelSpeed{//快慢速
            for section in sections {
                if section.sectionType == "normal" {
              
                    BFLog(message: "stuckPoints count is \(stuckPoints.count)")
                  
                    //一共裁剪的段数
                    var totalClipNum:Int = 0
                    for sticker in section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials() {
                        if sticker.type == StickerType.VIDEO.rawValue {
                            let asset: AVURLAsset = AVURLAsset(url: URL(fileURLWithPath: documensDirectory + sticker.locationPath), options: nil)
                        
                            var i:Int = 0
                            //一个视频切的的段落
                            var oneVideoClipNum:Int = 0
                         
                            var realUsedMusicDuration = 0.0
                            while realUsedMusicDuration <= CMTimeGetSeconds(asset.duration) {
                                //快速段
                                let LA = maxSpeed * (stuckPoints[i+1] - stuckPoints[i])
                                realUsedMusicDuration = realUsedMusicDuration + Float64(LA)
                                if(realUsedMusicDuration > CMTimeGetSeconds(asset.duration)){ break}
                                oneVideoClipNum = oneVideoClipNum + 1
                                //慢速段
                                let LB = minSpeed * (stuckPoints[i+2] - stuckPoints[i+1])
                                realUsedMusicDuration = realUsedMusicDuration + Float64(LB)
                                oneVideoClipNum = oneVideoClipNum + 1
                                i = i + 1
                            }
                            BFLog(message: "单个视频\(sticker.locationPath)时长：:\(CMTimeGetSeconds(asset.duration)) ,clipNum is:\(oneVideoClipNum)")
                        
                            for clipindex in 0 ... oneVideoClipNum - 1 {
                                // deep copy sticker model 防止只有一个对象
                                let stickerjson = sticker.toJSONString(prettyPrint: false)
                                let deepCopySticker = Mapper<PQEditVisionTrackMaterialsModel>().map(JSONString: stickerjson!)
                                // 设置循环模式和适配模式
                                deepCopySticker?.generateDefaultValues()
                                //当前分段的速度
                                let tempSpeed = (totalClipNum + clipindex) % 2 == 0 ? maxSpeed : minSpeed
                                if(totalClipNum + 1 + clipindex < stuckPoints.count){
                                    deepCopySticker?.speedRate = tempSpeed
                                    deepCopySticker?.model_in = Float64(tempSpeed) * Float64(stuckPoints[totalClipNum + clipindex] - (stuckPoints.first ?? 0.0))
                                    deepCopySticker?.out = Float64(tempSpeed) * Float64(stuckPoints[totalClipNum + 1 + clipindex] - (stuckPoints.first ?? 0.0))
                                    deepCopySticker?.timelineIn = Float64(stuckPoints[totalClipNum + clipindex])
                                    deepCopySticker?.timelineOut = Float64(stuckPoints[totalClipNum + 1 + clipindex])
                                }
                            
                                BFLog(message: " crilp is in \(deepCopySticker?.model_in ?? 0) out \(deepCopySticker?.out ?? 0) timelineIN:\(deepCopySticker?.timelineIn ?? 0) timelineOUT:\(deepCopySticker?.timelineOut ?? 0) 总时长\(CMTimeGetSeconds(asset.duration)) speedRate:\(String(describing: deepCopySticker?.speedRate))")

                                if deepCopySticker != nil {
                                    stickers.append(deepCopySticker!)
                                }
                              
                            }
                            totalClipNum = totalClipNum + oneVideoClipNum - 1
                        } else if sticker.type == StickerType.IMAGE.rawValue {
                            sticker.generateDefaultValues()
                            sticker.timelineIn = Float64(stuckPoints[totalClipNum])
                            sticker.timelineOut = Float64(stuckPoints[totalClipNum + 1])
                            stickers.append(sticker)
                        }
                    }
                }
            
            
            }
        } else if model == .createStickersModelOnlyMusic{ //仅音乐 首尾相连播放
            for section in sections {
                if section.sectionType == "normal" {

                    BFLog(message: "stuckPoints count is \(stuckPoints.count)")
                    //已经计算过所有贴纸的累计时长
                    var totalStickerTimer:Float64 = 0
                    for sticker in section.sectionTimeline!.visionTrack!.getEnableVisionTrackMaterials() {
                        if sticker.type == StickerType.VIDEO.rawValue {
                            let asset: AVURLAsset = AVURLAsset(url: URL(fileURLWithPath: documensDirectory + sticker.locationPath), options: nil)
     
                            BFLog(message: "单个视频\(sticker.locationPath)时长：:\(CMTimeGetSeconds(asset.duration))")
                            sticker.generateDefaultValues()
                            sticker.model_in = 0
                            sticker.out = CMTimeGetSeconds(asset.duration)

                            sticker.timelineIn = totalStickerTimer
                            sticker.timelineOut = sticker.timelineIn + sticker.out
                            
                            totalStickerTimer = totalStickerTimer +  sticker.out
                     
                   
                        } else if sticker.type == StickerType.IMAGE.rawValue {
                            sticker.generateDefaultValues()
                            sticker.timelineIn = totalStickerTimer
                            sticker.timelineOut =  sticker.timelineIn + sticker.out
                            
                            totalStickerTimer = totalStickerTimer +  sticker.out

                        }
                        
                        stickers.append(sticker)
                    }
                }
            
            
            }
        }

        return stickers
    }
    
    
    /// 根据档位取最后使用的卡点数据
    /// - Parameter seed: 档位速度
    /// - Returns: 最后使用的卡点
    func getUsedStuckPoint(seed:Int) -> Array<Float> {
        // 推荐卡点数
        var stuckPoints: Array = Array<Float>.init()

        var stuckPointsTemp = Array<Float>.init()
        for (index, dunshu) in stuckPointMusicData!.rhythmSdata[0].pointTimes.enumerated() {
            BFLog(message: "原所有卡点数:\(index) \(Float(dunshu) / Float(BASE_FILTER_TIMESCALE))")
            if Float64(dunshu) / Float64(BASE_FILTER_TIMESCALE) > (stuckPointMusicData?.startTime ?? 0) {
                stuckPointsTemp.append(Float(dunshu) / Float(BASE_FILTER_TIMESCALE))
            }
        }

        /*
         一，快慢速模式下取卡点 1 3 5
           快节奏为选中区域的所有点位，即0，1，2，3，4 5 6 7 8 9 10 ……
           适中为每两个点位取一个，即0，3，6，9 12
           慢节奏为每三个点位取一个，即0 5 10 15
         
         二，跳跃卡点模式下根据不同速度 取卡点 1,2,3
           快节奏为选中区域的所有点位，即0，1，2，3，4 5 6 7 8 9 10 ……
           适中为每两个点位取一个，即0，2，4，6……
           慢节奏为每三个点位取一个，即0，3，6，9……
         */
        BFLog(message: "stuckPointMusicData?.speed is \(String(describing: stuckPointMusicData?.speed))")
    
        //不丢
        if(seed == 1){
            stuckPoints = stuckPointsTemp
        }else{
            //根据档位要丢
            for (index, point) in stuckPointsTemp.enumerated() {
                if index % seed == 0 {
                    stuckPoints.append(point)
                }
            }
        }
  
        for point in stuckPoints {
            BFLog(message: "没有 start end 计算后的卡点数\(point)")
        }
        if stuckPoints.first != nil {
            stuckPoints.removeFirst()
        }
        if stuckPoints.last != nil {
            stuckPoints.removeLast()
        }
        // 开始时间是服务器返回， 结果时间根据策略计算的
        stuckPoints.insert(Float(stuckPointMusicData?.startTime ?? 0), at: 0)
        stuckPoints.insert(Float(stuckPointMusicData?.endTime ?? 0), at: stuckPoints.count)

        for point in stuckPoints {
            BFLog(message: "有 start end 计算后的卡点数\(point)")
        }

        BFLog(message: "stuckPoints count is \(stuckPoints.count)")
        
        return stuckPoints
        
        
         
    }
    
    /// 根据不同模式model，    maxSpeed ,minSpeed，      self?.stuckPointMusicData?.speed 档位，生成音乐时长和最终使用的卡点信息
    func dealParameter(model:createStickersModel){
 
        //清空上一次使用的卡点数据
        finallyStuckPoints.removeAll()
        //注意推荐时间位置和后面最近的卡点时间与0.3的关系
        //保存丢卡点处理后的卡点信息推荐开始到最后倒数第二个
        let stuckPointsTemp = getUsedStuckPoint(seed: stuckPointMusicData?.speed ?? 0)
        
        //最后真正要使用推荐结束时长
        var realUsedMusicDuration:Float = 0.0
        switch model {
        case .createStickersModelPoint://跳跃卡点
            // L/（n+1）  L -原视觉素材总时长  n-抛留倍数
            realUsedMusicDuration = Float(selectedTotalDuration) / (Float(maxSpeed) + 1)
            
            //要拼接的段数
            var clipNum:Int = 0
            //所有段的时长总和
            var tempTime:Float = 0.0
            var i:Int = 0
            while tempTime < Float(selectedTotalDuration) {
            
                //回环从头取
                if(i+2 > stuckPointsTemp.count){
                    i = 0
                }
                //快速段
                let LA =  (stuckPointsTemp[i+1] - stuckPointsTemp[i])
                tempTime = tempTime + Float(LA)
                if(tempTime > Float(selectedTotalDuration)){ break}
                //慢速段
                let LB = (stuckPointsTemp[i+2] - stuckPointsTemp[i+1])
                tempTime = tempTime + Float(LB)
                i = i + 1
                clipNum = clipNum + 1
            }
            
            //拼接要使用的卡点信息
            for i in 0...clipNum {
                finallyStuckPoints.append(stuckPointsTemp[i % stuckPointsTemp.count] + Float(i / stuckPointsTemp.count) * (stuckPointsTemp.first ?? 0))
            }
  
            break
        case .createStickersModelSpeed://快慢速

            /*
             - A-视频中的快速片段
             - B-视频中的慢速片段
             - d-在一档下音乐每个点位时长
             - n-不同音乐档位对应的d倍数，快节奏时，n=1；适中时，n=3；慢节奏时，n=5
             - L-原视觉素材时长
             - x-视频在A片段的播放倍速
             - y-视频在B片段的播放倍速
             */
            //LA=x*n*d，LB=y*n*d (n=1/3/5) 注：视频经过快慢速处理后的总时长约=L*2/(x+y)
            BFLog(message: "Ax快速为：\(maxSpeed) By慢速为：\(minSpeed) 档位 N为：\(stuckPointMusicData?.speed ?? 0)  使用的卡点总数:\(stuckPointsTemp.count)")
            
            //1,计算 AB 段的个数
            var i:Int = 0
            var LACount:Int = 0
            var LBCount:Int = 0
            while realUsedMusicDuration <= Float(selectedTotalDuration) {
            
                //回环从头取
                if(i+2 > stuckPointsTemp.count){
                    i = 0
                }
                //快速段
                let LA = maxSpeed * (stuckPointsTemp[i+1] - stuckPointsTemp[i])
                realUsedMusicDuration = realUsedMusicDuration + Float(LA)
                if(realUsedMusicDuration > Float(selectedTotalDuration)){ break}
                LACount = LACount + 1
                BFLog(message: "快速段段段段段段后LACount:\(LACount) realUsedMusicDuration:\(realUsedMusicDuration)")
                //慢速段
                let LB = minSpeed * (stuckPointsTemp[i+2] - stuckPointsTemp[i+1])
                realUsedMusicDuration = realUsedMusicDuration + Float(LB)
                LBCount = LBCount + 1
                BFLog(message: "慢速段段段段段段后LACount:\(LACount) realUsedMusicDuration:\(realUsedMusicDuration)")
                
                i = i + 1
            }
            
            //2，拼接要使用的卡点信息
            for (index,point) in stuckPointsTemp.enumerated() {
                 
                if(index < LACount + LBCount){
                    finallyStuckPoints.append(point)
                }
            }
            //2.1不足所有卡点回环拼接
            if( LACount + LBCount > stuckPointsTemp.count ){
        
                for (index,point) in stuckPointsTemp.enumerated() {
                    if(index <=  LACount + LBCount - stuckPointsTemp.count){
                        finallyStuckPoints.append(Float((stuckPointsTemp.last ?? 0) + point))
                    }
                }
            }
 
            BFLog(message: "计算后的\(realUsedMusicDuration) LACount:\(LACount) LBCount \(LBCount) 估算时长为：\(selectedTotalDuration * 2 / Float64(maxSpeed + minSpeed))")
       
            break
        case .createStickersModelOnlyMusic://仅音乐
            //时长为所有素材总时长
            finallyStuckPoints.append(Float(stuckPointMusicData?.startTime ?? 0) + Float(selectedTotalDuration))
         
            break
     
        }
        
        //计算最后使用的音频时长, 如果不用拼接音频时长度是卡点的倒数第二位时间
        let asset = AVURLAsset(url: URL(fileURLWithPath: documensDirectory + (stuckPointMusicData?.localPath ?? "")), options: nil)
 
        //原推荐卡点的倒数第二位时间
        let lastSecondPoint =   Float((stuckPointMusicData!.rhythmSdata[0].pointTimes[stuckPointMusicData!.rhythmSdata[0].pointTimes.count - 2] )) / Float(BASE_FILTER_TIMESCALE)
        
        finallyUserAudioTime = Float(lastSecondPoint)
        if(finallyStuckPoints.last ?? 0 > Float(CMTimeGetSeconds(asset.duration))){
           
            finallyUserAudioTime = Float(finallyStuckPoints.last ?? 0) + (Float(CMTimeGetSeconds(asset.duration)) - Float(lastSecondPoint))
            
        }
        
        playeTimeRange = CMTimeRange(start: CMTime(value: CMTimeValue( Float(stuckPointMusicData?.startTime ?? 0) * Float(BASE_FILTER_TIMESCALE)), timescale: BASE_FILTER_TIMESCALE), end: CMTime(value: CMTimeValue((Float(finallyStuckPoints.last ?? 0)) * Float(BASE_FILTER_TIMESCALE)), timescale: BASE_FILTER_TIMESCALE))
  
        // 更新裁剪时间条的的ui数据
        stuckPointCuttingView.videoDuration = CGFloat(finallyUserAudioTime)
        stuckPointCuttingView.updateEndTime(startTime:CGFloat(CMTimeGetSeconds( playeTimeRange.start)), endTime: CGFloat(  CMTimeGetSeconds( playeTimeRange.end)))
        
   
    }
}

// MARK: - 同步/下载素材相关

/// 同步/下载素材相关
extension PQStuckPointEditerController {
    /// 同步音乐相关数据
    /// - Returns: <#description#>
    func synchroMusicInfoData() {
        if (stuckPointMusicData?.rhythmSdata.count ?? 0) <= 0 {
            if synchroMarskView.superview == nil {
                UIApplication.shared.keyWindow?.addSubview(synchroMarskView)
            }
            PQStuckPointViewModel.stuckPointMusicDetailData(musicId: stuckPointMusicData?.musicId ?? "", originType: stuckPointMusicData?.originType ?? 1) { [weak self] newMusicData, _ in
                if newMusicData != nil, (newMusicData?.rhythmSdata.count ?? 0) > 0 {
                    self?.isStuckPointDataSuccess = true
                    self?.stuckPointMusicData?.rhythmSdata = newMusicData?.rhythmSdata ?? []
                    self?.stuckPointMusicData?.startTime = newMusicData?.startTime ?? 0
                    self?.stuckPointMusicData?.endTime = newMusicData?.endTime ?? 0
                    if newMusicData?.speed != nil {
                        self?.stuckPointMusicData?.speed = newMusicData?.speed ?? 2
                    }
                    if (self?.stuckPointMusicData?.rhythmSdata.count ?? 0) > 0 && (((self?.selectedDataCount ?? 0) - (self?.selectedImageDataCount ?? 0)) > 0 || (self?.selectedImageDataCount ?? 0) > 0 || (self?.selectedTotalDuration ?? 0) > 0) {
//                        self?.stuckPointMusicData?.endTime = (self?.stuckPointMusicData?.startTime ?? 0) + (self?.stuckPointMusicData?.stuckPointCuttingTime(videoCount: (self?.selectedDataCount ?? 0) - (self?.selectedImageDataCount ?? 0), imageCount: self?.selectedImageDataCount ?? 0, totalDuration: self?.selectedTotalDuration ?? 0) ?? 0)
                    }
                    self?.stuckPointCuttingView.updateEndTime(startTime: CGFloat(self?.stuckPointMusicData?.startTime ?? 0), endTime: CGFloat(self?.stuckPointMusicData?.endTime ?? 0))
                    if self?.stuckPointMusicData?.localPath == nil || (self?.stuckPointMusicData?.localPath?.count ?? 0) > 0 {
                        PQDownloadManager.downLoadFile(url: self?.stuckPointMusicData?.musicPath ?? "") { [weak self] filePath, error in
                            if error == nil, filePath != nil {
                                self?.isSynchroMusicInfoSuccess = true
                                self?.stuckPointMusicData?.localPath = filePath?.replacingOccurrences(of: documensDirectory, with: "")
                                // 处理所有数据完成
                                self?.dealWithDataSuccess()
                            } else {
                                if self?.synchroMarskView.superview != nil {
                                    self?.synchroMarskView.removeMarskView()
                                }
//                                PQUploadRemindView.showUploadRemindView(title: nil, attributedTitle: NSAttributedString(string: "加载音乐失败，请重新选择音乐"), summary: "", confirmTitle: nil) { [weak self] _, _ in
//                                    self?.navigationController?.popViewController(animated: true)
//                                }
                            }
                        }
                    } else {
                        self?.isSynchroMusicInfoSuccess = true
                        // 处理所有数据完成
                        self?.dealWithDataSuccess()
                    }
                    // 添加子视图
                    self?.addSubViews()
                } else {
                    if self?.synchroMarskView.superview != nil {
                        self?.synchroMarskView.removeMarskView()
                    }
//                    PQUploadRemindView.showUploadRemindView(title: nil, attributedTitle: NSAttributedString(string: "加载音乐失败，请重新选择音乐"), summary: "", confirmTitle: nil) { [weak self] _, _ in
//                        self?.navigationController?.popViewController(animated: true)
//                    }
                }
            }
        } else if stuckPointMusicData?.localPath == nil || (stuckPointMusicData?.localPath?.count ?? 0) > 0 {
            if synchroMarskView.superview == nil {
                UIApplication.shared.keyWindow?.addSubview(synchroMarskView)
            }
            isStuckPointDataSuccess = true
            PQDownloadManager.downLoadFile(url: stuckPointMusicData?.musicPath ?? "") { [weak self] filePath, error in
                if error == nil, filePath != nil {
                    self?.isSynchroMusicInfoSuccess = true
                    self?.stuckPointMusicData?.localPath = filePath?.replacingOccurrences(of: documensDirectory, with: "")
                    // 处理所有数据完成
                    self?.dealWithDataSuccess()
                } else {
                    if self?.synchroMarskView.superview != nil {
                        self?.synchroMarskView.removeMarskView()
                    }
//                    PQUploadRemindView.showUploadRemindView(title: nil, attributedTitle: NSAttributedString(string: "加载音乐失败，请重新选择音乐"), summary: "", confirmTitle: nil) { [weak self] _, _ in
//                        self?.navigationController?.popViewController(animated: true)
//                    }
                }
            }
        } else {
            isStuckPointDataSuccess = true
            // 处理所有数据完成
            dealWithDataSuccess()
        }
    }

    /// 导出相册数据
    /// - Returns: <#description#>
    func exportPhotoData() {
        // 取消所有的导出
        PQSingletoMemoryUtil.shared.allExportSession.forEach { _, exportSession in
            exportSession.cancelExport()
        }
        var isHaveVideo: Bool = false
        if selectedPhotoData != nil, (selectedPhotoData?.count ?? 0) > 0 {
            if synchroMarskView.superview == nil {
                UIApplication.shared.keyWindow?.addSubview(synchroMarskView)
            }
            let dispatchGroup = DispatchGroup()

            for photo in selectedPhotoData! {
                if photo.asset != nil, photo.asset?.mediaType == .video {
                    if !isHaveVideo {
                        isHaveVideo = true
                    }
                    dispatchGroup.enter()

                    PQPHAssetVideoParaseUtil.parasToAVAsset(phAsset: photo.asset!) { avAsset, _, _, _ in
                        if avAsset is AVURLAsset {
                            // 创建目录

                            let fileName = (avAsset as! AVURLAsset).url.absoluteString

                            BFLog(message: "video  fileName is\(fileName)")
                            let tempPhoto = self.selectedPhotoData?.first(where: { material in
                                material.asset == photo.asset
                            })

                            if (fileName.count) > 0 {
                                createDirectory(path: photoLibraryDirectory)
                                let outFilePath = photoLibraryDirectory + fileName.md5 + ".mp4"
                                // 文件存在先删除老文件
                                if FileManager.default.fileExists(atPath: outFilePath) {
                                    do {
                                        try FileManager.default.removeItem(at: NSURL.fileURL(withPath: outFilePath))
                                    } catch {
                                        BFLog(message: "导出相册视频-error == \(error)")
                                    }
                                }

                                do {
                                    try FileManager.default.copyItem(atPath: fileName.replacingOccurrences(of: "file:///", with: ""), toPath: outFilePath)
                                    print("Success to copy file.")
                                } catch {
                                    print("Failed to copy file.")
                                }

                                tempPhoto?.locationPath = outFilePath.replacingOccurrences(of: documensDirectory, with: "")
                                BFLog(message: "导出视频相册地址为 \(String(describing: tempPhoto?.locationPath))")

                                dispatchGroup.leave()
                            }
                        }
                    }
                }
            }

            dispatchGroup.notify(queue: DispatchQueue.main) { [weak self] in
                self?.isExportVideosSuccess = true
                BFLog(message: "所有相册视频导出成功")
                // 处理所有数据完成
                if isHaveVideo {
                    self?.dealWithDataSuccess()
                }
            }

            // 只有图片
            if !isHaveVideo {
                isExportVideosSuccess = true
                BFLog(message: "所有相册视频导出成功")

                dealWithDataSuccess()
            }
        }
    }

    /// 处理所有数据完成
    /// - Returns: <#description#>
    func dealWithDataSuccess() {
        BFLog(message: "三组参数：\(isSynchroMusicInfoSuccess)  \(isStuckPointDataSuccess) \(isExportVideosSuccess)")
        if !isSynchroMusicInfoSuccess || !isStuckPointDataSuccess || !isExportVideosSuccess {
            return
        }

        createPorjectData()
      
        BFLog(message: "界面编辑界面时参数 选择素材时长:\(selectedTotalDuration) 选择素材总数:\(selectedDataCount) 选择图片总数\(selectedImageDataCount)")
        
        
        //1 生成默认参数值
        /*
         - 当素材总时长∈[0-6)s 时，提示推荐仅配乐模式 or 快慢速模式
         - 当素材总时长∈[6-80)s 时，卡点抛留倍数为1x，即留1抛1
         - 当素材总时长∈[80,120)s 时，卡点抛留倍数为2x，即留1抛2
         - 当素材总时长∈[120,160)s 时，卡点抛留倍数为3x，即留1抛3
         - 当素材总时长∈[160,200)s 时，卡点抛留倍数为4x，即留1抛4
         - 当素材总时长∈[200,240)s 时，卡点抛留倍数为5x，即留1抛5
         - 当素材总时长∈[240,∞)s 时，卡点抛留倍数为5x，即留1抛5
         */
        //1.1) 生成跳跃卡点的默认值
        if(selectedTotalDuration >= 6 && selectedTotalDuration < 80){
            lastJumpSpeedSelectIndex = 0
        }else if(selectedTotalDuration >= 80 && selectedTotalDuration < 120){
            lastJumpSpeedSelectIndex = 1
        }else if(selectedTotalDuration >= 120 && selectedTotalDuration < 160){
            lastJumpSpeedSelectIndex = 2
        }else if(selectedTotalDuration >= 160 && selectedTotalDuration < 200){
            lastJumpSpeedSelectIndex = 3
        }else if(selectedTotalDuration >= 200 && selectedTotalDuration < 240){
            lastJumpSpeedSelectIndex = 4
        }else if(selectedTotalDuration >= 240){
            lastJumpSpeedSelectIndex = 5
        }
        
        /*默认进入快慢速模式
         - 当素材总时长∈[120,144]s 时，快慢速处理方式：快速为 6x，慢速为 1.2x，效果是快&快
         - 当素材总时长∈[70,120)s 时，快慢速处理方式：快速为5x，慢速为 1x，效果是快&正常
         - 当素材总时长∈[56,70)s 时，快慢速处理方式：快速为3x，慢速为 0.5x，效果是快&慢
         - 当素材总时长∈[17.5,56)s 时，快慢速处理方式：快速为 2.4x，慢速为 0.4x，效果是快&慢
         - 当素材总时长∈[10.5,17.5)s 时，快慢速处理方式：快速为 1.8x，慢速为 0.3x，效果是快&慢
         - 当素材总时长∈(0,10.5)s 时，快慢速处理方式：快速为 1x，慢速为 0.2x，效果是正常&慢
         
         */
        //1.2)生成快慢速的默认值
        if(selectedTotalDuration >= 120 && selectedTotalDuration <= 144){
            lastSpeedSelectIndex = 0
        }else if(selectedTotalDuration >= 70 && selectedTotalDuration < 120){
            lastSpeedSelectIndex = 1
        }else if(selectedTotalDuration >= 56 && selectedTotalDuration < 70){
            lastSpeedSelectIndex = 2
        }else if(selectedTotalDuration >= 17.5 && selectedTotalDuration < 56){
            lastSpeedSelectIndex = 3
        }else if(selectedTotalDuration >= 10.5 && selectedTotalDuration < 17.5){
            lastSpeedSelectIndex = 4
        }else if(selectedTotalDuration > 0 && selectedTotalDuration < 10.5){
            lastSpeedSelectIndex = 5
        }
        
        
        /*
          文档规则 https://w42nne6hzg.feishu.cn/docs/doccnQZm1uCfkU4QtJb5fLxYk4d#
         */
        //2,根据所选择所有素材时长判断哪些功能不可能用，及进入默认模式
        //全是图片
        if(selectedDataCount == selectedImageDataCount){
            BFLog(message: "全是图片 \(selectedDataCount) \(selectedImageDataCount)")
            speedStuckBtn.isEnabled = false
            //默认进入跳跃卡点模式
            editModelClick(sender: jumpPointBtn)
            speedSettingView.setSelectItem(index: lastJumpSpeedSelectIndex)
            
        }else{
            
            //1,根据素材输入确定可用模式及推荐模式
            //1)0-6s时卡点不能使用， 纯图片时快慢速不可用
            if(selectedTotalDuration > 0 && selectedTotalDuration < 6){
                jumpPointBtn.isEnabled = false
            }
            //默认进入快慢速模式
            if(selectedTotalDuration > 0 && selectedTotalDuration <= 144){
                
                editModelClick(sender: speedStuckBtn)
                speedSettingView.setSelectItem(index: lastSpeedSelectIndex)
     
            }else{
                //默认进入卡点模式
                editModelClick(sender: jumpPointBtn)
                speedSettingView.setSelectItem(index: lastJumpSpeedSelectIndex)
        
            }
        }
 
    }
}
